/*
Copyright (C) 2003-2006 Andrey Nazarov

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "script_local.h"

typedef struct {
	char			*name;
	evalType_t		type;
	int				sizeRank;
} type_t;

type_t simpleTypes[] = {
	{ "void",	EV_VOID			},
	{ "int",	EV_INT,		1	},
	{ "float",	EV_FLOAT,	2	},
	{ "string",	EV_STRING,	3	},
	{ "bool",	EV_BOOL			},
	{ "vector",	EV_VECTOR		},
	{ "color",	EV_COLOR		},
	{ "rect",	EV_RECT			},

	{ "cvar",		EV_CVAR			},
	{ "macro",		EV_MACRO		},
	{ "function",	EV_FUNCTION		},
	{ "field",		EV_FIELD		},

	{ NULL }
};

scope_t	globalScope;


list_t tempdefs;


def_t *currentFunction;
scope_t *currentScope;
scope_t *scopes;

#define MAX_STATEMENTS	1024

static statement_t	statements[MAX_STATEMENTS];

statementBuffer_t	stbuffer = { statements, MAX_STATEMENTS };
statementBuffer_t	*stbuf;

void Script_ParseDef( void );


#define RIGHT_FLAG	16
#define UNARY_FLAG	32

#define PRIORITY_TOP	6
#define PRIORITY_STORE	( 5 | RIGHT_FLAG )
#define PRIORITY_INCDEC	( 1 | UNARY_FLAG | RIGHT_FLAG )
#define PRIORITY_NOT	4
#define PRIORITY_CALL	1
#define PRIORITY_TERM	0

#define OP( x )		"OP_"#x, OP_##x

operator_t	opcodes[] = {
	{ NULL,		OP( BREAK ) },
	{ NULL,		OP( NOP ) },

	{ ".",		OP( INDIRECT ),	1,				EV_COLOR,	EV_FIELD,	EV_FLOAT },
	{ ".",		OP( INDIRECT ),	1,				EV_RECT,	EV_FIELD,	EV_INT },
//	{ ".",		OP( INDIRECT ),	1,				EV_STRING,	EV_FIELD,	EV_INT },
//	{ ".",		OP( INDIRECT_P ),	1,			EV_CVAR,	EV_FIELD,	EV_INT },

	{ "*",		OP( MUL_F ),	2,				EV_FLOAT,	EV_FLOAT,	EV_FLOAT },
	{ "*",		OP( MUL_I ),	2,				EV_INT,		EV_INT,		EV_INT },

	{ "/",		OP( DIV_F ),	2,				EV_FLOAT,	EV_FLOAT,	EV_FLOAT },
	{ "/",		OP( DIV_I ),	2,				EV_INT,		EV_INT,		EV_INT },

	{ "%",		OP( MOD_I ),	2,				EV_INT,		EV_INT,		EV_INT },

	{ "+",		OP( ADD_F ),	3,				EV_FLOAT,	EV_FLOAT,	EV_FLOAT },
	{ "+",		OP( ADD_I ),	3,				EV_INT,		EV_INT,		EV_INT },
	{ "+",		OP( ADD_V ),	3,				EV_VECTOR,	EV_VECTOR,	EV_VECTOR },
	{ "+",		OP( ADD_C ),	3,				EV_COLOR,	EV_COLOR,	EV_COLOR },
	{ "+",		OP( ADD_S ),	3,				EV_STRING,	EV_STRING,	EV_STRING },

	{ "-",		OP( SUB_F ),	3,				EV_FLOAT,	EV_FLOAT,	EV_FLOAT },
	{ "-",		OP( SUB_I ),	3,				EV_INT,		EV_INT,		EV_INT },
	{ "-",		OP( SUB_V ),	3,				EV_VECTOR,	EV_VECTOR,	EV_VECTOR },
	{ "-",		OP( SUB_C ),	3,				EV_COLOR,	EV_COLOR,	EV_COLOR },

	{ "=",		OP( STORE_F ),	PRIORITY_STORE,				EV_FLOAT,			EV_FLOAT	},
	{ "=",		OP( STORE_I ),	PRIORITY_STORE,				EV_INT,				EV_INT		},
	{ "=",		OP( STORE_V ),	PRIORITY_STORE,				EV_VECTOR,			EV_VECTOR	},
	{ "=",		OP( STORE_C ),	PRIORITY_STORE,				EV_COLOR,			EV_COLOR	},
	{ "=",		OP( STORE_R ),	PRIORITY_STORE,				EV_RECT,			EV_RECT		},
	{ "=",		OP( STORE_S ),	PRIORITY_STORE,				EV_STRING,			EV_STRING	},
	{ "=",		OP( STORE_B ),	PRIORITY_STORE,				EV_BOOL,			EV_BOOL		},

	{ "*=",		OP( UMUL_F ),	PRIORITY_STORE,				EV_FLOAT,			EV_FLOAT	},
	{ "*=",		OP( UMUL_I ),	PRIORITY_STORE,				EV_INT,				EV_INT		},

	{ "/=",		OP( UDIV_F ),	PRIORITY_STORE,				EV_FLOAT,			EV_FLOAT	},
	{ "/=",		OP( UDIV_I ),	PRIORITY_STORE,				EV_INT,				EV_INT		},

	{ "+=",		OP( UADD_F ),	PRIORITY_STORE,				EV_FLOAT,			EV_FLOAT 	},
	{ "+=",		OP( UADD_I ),	PRIORITY_STORE,				EV_INT,				EV_INT 		},
	{ "+=",		OP( UADD_V ),	PRIORITY_STORE,				EV_VECTOR,			EV_VECTOR	},
	{ "+=",		OP( UADD_C ),	PRIORITY_STORE,				EV_COLOR,			EV_COLOR	},
	{ "+=",		OP( UADD_S ),	PRIORITY_STORE,				EV_STRING,			EV_STRING	},

	{ "-=",		OP( USUB_F ),	PRIORITY_STORE,				EV_FLOAT,			EV_FLOAT	},
	{ "-=",		OP( USUB_I ),	PRIORITY_STORE,				EV_INT,				EV_INT		},
	{ "-=",		OP( USUB_V ),	PRIORITY_STORE,				EV_VECTOR,			EV_VECTOR	},
	{ "-=",		OP( USUB_C ),	PRIORITY_STORE,				EV_COLOR,			EV_COLOR	},

	{ "++",		OP( UINC_F ),	PRIORITY_INCDEC,			EV_FLOAT },
	{ "++",		OP( UINC_I ),	PRIORITY_INCDEC,			EV_INT },

	{ "--",		OP( UDEC_F ),	PRIORITY_INCDEC,			EV_FLOAT  },
	{ "--",		OP( UDEC_I ),	PRIORITY_INCDEC,			EV_INT  },

	{ "==",		OP( EQ_F ),		4,				EV_FLOAT,	EV_FLOAT,	EV_BOOL	},
	{ "==",		OP( EQ_I ),		4,				EV_INT,		EV_INT,		EV_BOOL	},
	{ "==",		OP( EQ_I ),		4,				EV_BOOL,	EV_BOOL,	EV_BOOL	},
	{ "==",		OP( EQ_V ),		4,				EV_VECTOR,	EV_VECTOR,	EV_BOOL	},
	{ "==",		OP( EQ_C ),		4,				EV_COLOR,	EV_COLOR,	EV_BOOL	},
	{ "==",		OP( EQ_S ),		4,				EV_STRING,	EV_STRING,	EV_BOOL	},

	{ "!=",		OP( NE_F ),		4,				EV_FLOAT,	EV_FLOAT,	EV_BOOL },
	{ "!=",		OP( NE_I ),		4,				EV_INT,		EV_INT,		EV_BOOL },
	{ "!=",		OP( NE_I ),		4,				EV_BOOL,	EV_BOOL,	EV_BOOL },
	{ "!=",		OP( NE_V ),		4,				EV_VECTOR,	EV_VECTOR,	EV_BOOL	},
	{ "!=",		OP( NE_C ),		4,				EV_COLOR,	EV_COLOR,	EV_BOOL	},
	{ "!=",		OP( NE_S ),		4,				EV_STRING,	EV_STRING,	EV_BOOL	},

	{ ">=",		OP( GE_F ),		4,				EV_FLOAT,	EV_FLOAT,	EV_BOOL },
	{ ">=",		OP( GE_I ),		4,				EV_INT,		EV_INT,		EV_BOOL },
	{ ">=",		OP( GE_S ),		4,				EV_STRING,	EV_STRING,	EV_BOOL },

	{ ">",		OP( GT_F ),		4,				EV_FLOAT,	EV_FLOAT,	EV_BOOL },
	{ ">",		OP( GT_I ),		4,				EV_INT,		EV_INT,		EV_BOOL },
	{ ">",		OP( GT_S ),		4,				EV_STRING,	EV_STRING,	EV_BOOL },

	{ "<=",		OP( LE_F ),		4,				EV_FLOAT,	EV_FLOAT,	EV_BOOL },
	{ "<=",		OP( LE_I ),		4,				EV_INT,		EV_INT,		EV_BOOL },
	{ "<=",		OP( LE_S ),		4,				EV_STRING,	EV_STRING,	EV_BOOL },

	{ "<",		OP( LT_F ),		4,				EV_FLOAT,	EV_FLOAT,	EV_BOOL },
	{ "<",		OP( LT_I ),		4,				EV_INT,		EV_INT,		EV_BOOL },
	{ "<",		OP( LT_S ),		4,				EV_STRING,	EV_STRING,	EV_BOOL },

	{ "|",		OP( BITOR ),	3,				EV_INT,	EV_INT,	EV_INT },
	{ "&",		OP( BITAND ),	3,				EV_INT,	EV_INT,	EV_INT },
	{ "^",		OP( BITXOR ),	3,				EV_INT,	EV_INT,	EV_INT },

	{ "|=",		OP( UBITOR ),	PRIORITY_STORE,				EV_INT,	EV_INT },
	{ "&=",		OP( UBITAND ),	PRIORITY_STORE,				EV_INT,	EV_INT },
	{ "^=",		OP( UBITXOR ),	PRIORITY_STORE,				EV_INT,	EV_INT },

	{ NULL,		OP( NOT ),		-1,	EV_BOOL,	EV_VOID,	EV_BOOL },
	{ NULL,		OP( NEG_F ),	-1,	EV_FLOAT,	EV_VOID,	EV_FLOAT },
	{ NULL,		OP( NEG_I ),	-1,	EV_INT,		EV_VOID,	EV_INT },
	{ NULL,		OP( BITINV ),	-1,	EV_INT,		EV_VOID,	EV_INT },

	{ NULL,		OP( LOADP_F )	},
	{ NULL,		OP( LOADP_I )	},
	{ NULL,		OP( LOADP_V )	},
	{ NULL,		OP( LOADP_C )	},
	{ NULL,		OP( LOADP_S )	},
	{ NULL,		OP( LOADP_B )	},

	{ NULL,		OP( STOREP_F )	},
	{ NULL,		OP( STOREP_I )	},
	{ NULL,		OP( STOREP_V )	},
	{ NULL,		OP( STOREP_C )	},
	{ NULL,		OP( STOREP_R )	},
	{ NULL,		OP( STOREP_S )	},
	{ NULL,		OP( STOREP_B )	},

	{ NULL,		OP( ENTER )	},
	{ NULL,		OP( LEAVE )	},
	{ NULL,		OP( CALL )	},
	{ NULL,		OP( JUMP )	},

	{ NULL,		OP( IFNOT )	},

	{ "<CONV>",		OP( CV_FI ), -1, EV_FLOAT,	EV_VOID,		EV_INT		},
	{ "<CONV>",		OP( CV_IF ), -1, EV_INT,	EV_VOID,		EV_FLOAT	},
	{ "<CONV>",		OP( CV_SI ), -1, EV_STRING,	EV_VOID,		EV_INT		},
	{ "<CONV>",		OP( CV_IS ), -1, EV_INT,	EV_VOID,		EV_STRING	},
	{ "<CONV>",		OP( CV_SF ), -1, EV_STRING,	EV_VOID,		EV_FLOAT	},
	{ "<CONV>",		OP( CV_FS ), -1, EV_FLOAT,	EV_VOID,		EV_STRING	},
	{ "<CONV>",		OP( CV_CS ), -1, EV_COLOR,	EV_VOID,		EV_STRING	},
	{ "<CONV>",		OP( CV_BS ), -1, EV_BOOL,	EV_VOID,		EV_STRING	},
	{ "<CONV>",		OP( CV_CVI ), -1, EV_CVAR,	EV_VOID,		EV_INT		},
	{ "<CONV>",		OP( CV_CVF ), -1, EV_CVAR,	EV_VOID,		EV_FLOAT	},
	{ "<CONV>",		OP( CV_CVS ), -1, EV_CVAR,	EV_VOID,		EV_STRING	},
	{ "<CONV>",		OP( CV_MI ), -1, EV_MACRO,	EV_VOID,		EV_INT		},
	{ "<CONV>",		OP( CV_MF ), -1, EV_MACRO,	EV_VOID,		EV_FLOAT	},
	{ "<CONV>",		OP( CV_MS ), -1, EV_MACRO,	EV_VOID,		EV_STRING	},

	{ NULL }
};

#undef OP

void Script_EnterScope( void ) {
	scope_t *scope;

	scope = Z_Malloc( sizeof( *scope ) );
	scope->parent = currentScope;
	currentScope = scope;
	scope->next = scopes;
	scopes = scope;
}

void Script_LeaveScope( void ) {
	currentScope = currentScope->parent;
	if( !currentScope ) {
		Lex_Error( "Script_LeaveScope: underflow" );
	}
}

qboolean Script_IsSimpleType( void ) {
	type_t *type;

	if( lex.tokenType != TT_NAME ) {
		return qfalse;
	}

	for( type = simpleTypes; type->name; type++ ) {
		if( type->type > LAST_SIMPLE_TYPE ) {
			break;
		}
		if( !strcmp( type->name, lex.token ) ) {
			return qtrue;
		}
	}
	
	return qfalse;
}

/*
================
Script_ParseSimpleType
================
*/
evalType_t Script_ParseSimpleType( void ) {
	type_t *type;

	if( lex.tokenType != TT_NAME ) {
		Lex_Error( "Expected type or name, found '%s'", lex.token );
	}

	for( type = simpleTypes; type->name; type++ ) {
		if( type->type > LAST_SIMPLE_TYPE ) {
			break;
		}
		if( !strcmp( type->name, lex.token ) ) {
			Lex();
			return type->type;
		}
	}
	
	Lex_Error( "Expected type, found '%s'", lex.token );
	return 0;
	
}

/*
================
Script_ParseName
================
*/
char *Script_ParseName( void ) {
	static char name[MAX_STRING_CHARS];

	if( lex.tokenType != TT_NAME ) {
		Lex_Error( "Expected name, found '%s'", lex.token );
	}

	if( lex.tokenLength > MAX_STRING_CHARS - 1 ) {
		Lex_Error( "Name too long" );
	}

	strcpy( name, lex.token );

	Lex();

	return name;
}

def_t *Script_FindLocalDef( char *name, scope_t *scope ) {
	listElem_t *elem;

	for( elem = scope->defs.first; elem; elem = elem->next ) {
		if( !strcmp( (( def_t * )elem)->name, name ) ) {
			return ( def_t * )elem;
		}
	}

	return NULL;
}

void Script_FreeDef( def_t *def ) {
	if( def->type == EV_STRING && !( def->flags & DEF_DONTFREE ) ) {
		Z_Free( def->eval.string );
	}
	List_DeleteElem( def );
	Z_Free( def );
}


def_t *Script_FindDef( char *name ) {
	scope_t *scope;
	def_t *def;

	for( scope = currentScope; scope; scope = scope->parent ) {
		def = Script_FindLocalDef( name, scope );
		if( def ) {
			return def;
		}
	} 

	return NULL;
}

/* creates named def, visible to the code */
def_t *Script_CreateDef( const char *name, evalType_t type ) {
	def_t *def;
	int length;

	length = strlen( name );
	def = Z_Malloc( sizeof( *def ) + length );
	strcpy( def->name, name );
	def->type = type;

	if( type == EV_STRING ) {
		def->eval.string = Z_Malloc( MAX_STRING_CHARS );
	}

	List_Append( &currentScope->defs, def );

	return def;
}

/* creates temporary def for storing intermediate results */
def_t *Script_CreateTempDef( evalType_t type ) {
	def_t *def;

	def = Z_Malloc( sizeof( *def ) );
	def->name[0] = 0;
	def->type = type;
	def->flags = DEF_TEMP;

	if( type == EV_STRING ) {
		def->eval.string = Z_Malloc( MAX_STRING_CHARS );
	}

	List_Append( &tempdefs, def );

	return def;
}

/* does not allocate buffer for strings */
def_t *Script_CreateTempConstDef( evalType_t type ) {
	def_t *def;

	def = Z_Malloc( sizeof( *def ) );
	def->name[0] = 0;
	def->type = type;
	def->flags = DEF_TEMP|DEF_CONST;

	List_Append( &tempdefs, def );

	return def;
}

void Script_Free( void ) {
	scope_t *scope, *nextScope;
	listElem_t *elem, *next;
	def_t *def;

	for( scope = scopes; scope; scope = nextScope ) {
		nextScope = scope->next;
		for( elem = scope->defs.first; elem; elem = next ) {
			next = elem->next;
			def = ( def_t * )elem;
			if( def->type == EV_STRING && !( def->flags & DEF_DONTFREE ) ) {
				Z_Free( def->eval.string );
			}
			Z_Free( def );
		}
		Z_Free( scope );
	}
	

	for( elem = tempdefs.first; elem; elem = next ) {
		next = elem->next;
		def = ( def_t * )elem;
		if( def->type == EV_STRING && !( def->flags & DEF_DONTFREE ) ) {
			Z_Free( def->eval.string );
		}
		Z_Free( def );
	}

	stbuffer.numStatements = 0;
	memset( statements, 0, sizeof( statements ) );

	scopes = NULL;
	List_Clear( &tempdefs );

	memset( &globalScope, 0, sizeof( globalScope ) );
	currentFunction = NULL;
}

/*
================
Script_ParseParameters
================
*/
void Script_ParseParameters( void ) {
	evalType_t type;
	char *name;
	def_t *def;
	int numArgs;

	if( Lex_Check( ")" ) ) {
		return;
	}

	if( Lex_Check( "void" ) ) {
		Lex_Expect( ")" );
		return;
	}

	numArgs = 0;
	do {
		if( numArgs == MAX_ARGS ) {
			Lex_Error( "'%s' has too many parameters", currentFunction->name );
		}
		type = Script_ParseSimpleType();
		name = Script_ParseName();
		if( Script_FindLocalDef( name, currentScope ) ) {
			Lex_Error( "Redefenition of '%s'", name );
		}
		def = Script_CreateDef( name, type );
		currentFunction->args[numArgs++] = def;
	} while( Lex_Check( "," ) );

	Lex_Expect( ")" );

	currentFunction->numArgs = numArgs;
}

def_t *Script_Expression( int priority );


operator_t *Script_FindOpcode( int code ) {
	operator_t *op;

	for( op = opcodes; op->name; op++ ) {
		if( op->code == code ) {
			return op;
		}
	}

	Com_Error( ERR_DROP, "Script_FindOpcode: %i not found", code );

	return NULL;
}

statement_t *Script_AllocStatement( void ) {
	statement_t *s;

	if( stbuf->numStatements == stbuf->maxStatements ) {
		Com_Error( ERR_DROP, "MAX_STATEMENTS exceeded" );
	}

	s = &stbuf->statements[stbuf->numStatements++];
	return s;
}

def_t *Script_Load( def_t *def ) {
	statement_t *s;
	def_t *tmp;

	tmp = Script_CreateTempConstDef( def->type );

	s = Script_AllocStatement();
	switch( def->type ) {
	case EV_INT:
		s->op = OP_LOADP_I;
		break;
	case EV_FLOAT:
		s->op = OP_LOADP_F;
		break;
	case EV_VECTOR:
		s->op = OP_LOADP_V;
		break;
	case EV_COLOR:
		s->op = OP_LOADP_C;
		break;
	case EV_STRING:
		/* simply copies string pointer */
		s->op = OP_LOADP_S;
		tmp->flags |= DEF_DONTFREE;
		break;
	default:
		Com_Error( ERR_DROP, "Script_Load: bad def type" );
		break;
	}
	
	s->a = def;
	s->c = tmp;
	
	return tmp;
}

/*
================
Script_Optimize
================
*/
def_t *Script_Optimize( operator_t *op, def_t *a, def_t *b ) {
	def_t *c;
	int length;
	char buffer[MAX_STRING_CHARS];
	statement_t *s;

	if( !( a->flags & DEF_IMMEDIATE ) ) {
		return NULL;
	}
	if( a->flags & DEF_POINTER ) {
		Lex_Error( "DEF_IMMEDIATE with DEF_POINTER set" );
	}

	if( b ) {
		if( !( b->flags & DEF_IMMEDIATE ) ) {
			return NULL;
		}
		if( b->flags & DEF_POINTER ) {
			Lex_Error( "DEF_IMMEDIATE with DEF_POINTER set" );
		}
	}

	c = Script_CreateTempConstDef( op->c );
	c->flags |= DEF_IMMEDIATE;

	switch( op->code ) {
	case OP_MUL_F:
		c->eval.value = a->eval.value * b->eval.value;
		break;
	case OP_DIV_F:
		if( b->eval.value == 0 ) {
			Lex_Error( "Division by zero" );
		}
		c->eval.value = a->eval.value / b->eval.value;
		break;
	case OP_ADD_F:
		c->eval.value = a->eval.value + b->eval.value;
		break;
	case OP_SUB_F:
		c->eval.value = a->eval.value - b->eval.value;
		break;

	case OP_EQ_F:
		c->eval.integer = a->eval.value == b->eval.value;
		break;
	case OP_NE_F:
		c->eval.integer = a->eval.value != b->eval.value;
		break;
	case OP_GE_F:
		c->eval.integer = a->eval.value >= b->eval.value;
		break;
	case OP_GT_F:
		c->eval.integer = a->eval.value > b->eval.value;
		break;
	case OP_LE_F:
		c->eval.integer = a->eval.value <= b->eval.value;
		break;
	case OP_LT_F:
		c->eval.integer = a->eval.value < b->eval.value;
		break;

	case OP_NEG_F:
		c->eval.value = -a->eval.value;	
		break;

	case OP_MUL_I:
		c->eval.integer = a->eval.integer * b->eval.integer;
		break;
	case OP_DIV_I:
		if( b->eval.integer == 0 ) {
			Lex_Error( "Division by zero" );
		}
		c->eval.integer = a->eval.integer / b->eval.integer;
		break;
	case OP_MOD_I:
		if( b->eval.integer == 0 ) {
			Lex_Error( "Division by zero" );
		}
		c->eval.integer = a->eval.integer % b->eval.integer;
		break;
	case OP_ADD_I:
		c->eval.integer = a->eval.integer + b->eval.integer;
		break;
	case OP_SUB_I:
		c->eval.integer = a->eval.integer - b->eval.integer;
		break;

	case OP_EQ_I:
		c->eval.integer = a->eval.integer == b->eval.integer;
		break;
	case OP_NE_I:
		c->eval.integer = a->eval.integer != b->eval.integer;
		break;
	case OP_GE_I:
		c->eval.integer = a->eval.integer >= b->eval.integer;
		break;
	case OP_GT_I:
		c->eval.integer = a->eval.integer > b->eval.integer;
		break;
	case OP_LE_I:
		c->eval.integer = a->eval.integer <= b->eval.integer;
		break;
	case OP_LT_I:
		c->eval.integer = a->eval.integer < b->eval.integer;
		break;

	case OP_BITOR:
		c->eval.integer = a->eval.integer | b->eval.integer;
		break;
	case OP_BITAND:
		c->eval.integer = a->eval.integer & b->eval.integer;
		break;
	case OP_BITXOR:
		c->eval.integer = a->eval.integer ^ b->eval.integer;
		break;

	case OP_NEG_I:
		c->eval.integer = -a->eval.integer;
		break;
	case OP_BITINV:
		c->eval.integer = ~a->eval.integer;
		break;

	case OP_ADD_S:
		length = strlen( a->eval.string ) + strlen( b->eval.string ) + 1;
		if( length > MAX_STRING_CHARS ) {
			Lex_Error( "String constant exceedes MAX_STRING_CHARS" );
		}
		c->eval.string = Z_Malloc( length );
		strcpy( c->eval.string, a->eval.string );
		strcat( c->eval.string, b->eval.string );
		break;

	case OP_EQ_S:
		c->eval.integer = strcmp( a->eval.string, b->eval.string ) == 0;
		break;
	case OP_NE_S:
		c->eval.integer = strcmp( a->eval.string, b->eval.string ) != 0;
		break;
	case OP_GE_S:
		c->eval.integer = strcmp( a->eval.string, b->eval.string ) >= 0;
		break;
	case OP_GT_S:
		c->eval.integer = strcmp( a->eval.string, b->eval.string ) > 0;
		break;
	case OP_LE_S:
		c->eval.integer = strcmp( a->eval.string, b->eval.string ) <= 0;
		break;
	case OP_LT_S:
		c->eval.integer = strcmp( a->eval.string, b->eval.string ) < 0;
		break;
	case OP_CV_FI:
		c->eval.integer = a->eval.value;
		break;
	case OP_CV_IF:
		c->eval.value = a->eval.integer;
		break;
	case OP_CV_SI:
		c->eval.integer = atoi( a->eval.string );
		break;
	case OP_CV_IS:
		Com_sprintf( buffer, sizeof( buffer ), "%d", a->eval.integer );
		c->eval.string = Com_CopyString( buffer );
		break;
	case OP_CV_SF:
		c->eval.value = atof( a->eval.string );
		break;
	case OP_CV_FS:
		if( a->eval.value == ( int )a->eval.value ) {
			Com_sprintf( buffer, sizeof( buffer ), "%d", ( int )a->eval.value );
		} else {
			Com_sprintf( buffer, sizeof( buffer ), "%f", a->eval.value );
		}
		c->eval.string = Com_CopyString( buffer );
		break;
	case OP_CV_CS:
		Com_sprintf( buffer, sizeof( buffer ), "(%f %f %f %f)",
			a->eval.color[0], a->eval.color[1], a->eval.color[2], a->eval.color[3] );
		c->eval.string = Com_CopyString( buffer );
		break;
	case OP_CV_BS:
		if( a->eval.integer ) {
			c->eval.string = "true";
		} else {
			c->eval.string = "false";
		}
		c->flags |= DEF_DONTFREE;
		break;
	case OP_IFNOT:
		s = Script_AllocStatement();
		if( !a->eval.integer ) {
			s->op = OP_JUMP;
		} else {
			s->op = OP_NOP;
		}
		break;
	case OP_NOT:
		c->eval.integer = !a->eval.integer;
		break;
	default:
		Lex_Error( "Script_Optimize: unsupported opcode: %d", op->code );
		break;
	}

	/* free a and b */
	if( !( a->flags & DEF_DONTFREE ) ) {
		Script_FreeDef( a );
	}
	if( b && !( b->flags & DEF_DONTFREE ) ) {
		Script_FreeDef( b );
	}
	
	return c;
}


/*
================
Script_Statement
================
*/
def_t *Script_Statement( operator_t *op, def_t *a, def_t *b ) {
	statement_t *s;
	def_t *c;

	c = Script_Optimize( op, a, b );
	if( c ) {
		return c;
	}

	c = Script_CreateTempDef( op->c );

	if( op->string && op->string[0] == '.' ) {
		/* indirection */
		c->flags |= DEF_POINTER;
	}
	if( op->code == OP_CV_CVS ) {
		/* cvar to string conversion */
		c->flags |= DEF_CONST|DEF_DONTFREE;
	}

	if( a->flags & DEF_POINTER ) {
		a = Script_Load( a );
	}
	if( b && ( b->flags & DEF_POINTER ) ) {
		b = Script_Load( b );
	}

	s = Script_AllocStatement();
	s->op = op->code;
	s->a = a;
	s->b = b;
	s->c = c;

	return c;

}

/*
================
Script_RightStatement

Used for assignment operators only
================
*/
def_t *Script_RightStatement( operator_t *op, def_t *a, def_t *b ) {
	statement_t *s;

	if( a->flags & DEF_IMMEDIATE ) {
		Lex_Error( "Assignment to immediate" );
	}

	if( a->flags & DEF_CONST ) {
		Lex_Error( "Assignment to const" );
	}

	if( b && ( b->flags & DEF_POINTER ) ) {
		b = Script_Load( b );
	}

	s = Script_AllocStatement();
	if( a->flags & DEF_POINTER ) {
		if( op->code < OP_STORE_I || op->code > OP_STORE_P ) {
			Lex_Error( "Operator can't assign to pointers" );
		}
		s->op = op->code - OP_STORE_I + OP_STOREP_I;
	} else {
		s->op = op->code;
	}
	s->a = b;
	s->b = a;
	s->c = NULL;

	return a;

}


/*
================
Script_ParseValue
================
*/
def_t *Script_ParseValue( void ) {
	def_t *def;
	char *name;
	xmacro_t f;

	if( lex.tokenType == TT_IMMEDIATE ) {
		switch( lex_immediateType ) {
		case IT_INT:
			def = Script_CreateTempConstDef( EV_INT );
			def->eval.integer = lex_immediate.integer;
			break;
		case IT_FLOAT:
			def = Script_CreateTempConstDef( EV_FLOAT );
			def->eval.value = lex_immediate.value;
			break;
		case IT_STRING:
			def = Script_CreateTempConstDef( EV_STRING );
			def->eval.string = Com_CopyString( lex_immediate.string );
			break;
		case IT_VECTOR3:
			def = Script_CreateTempConstDef( EV_VECTOR );
			def->eval.vector[0] = lex_immediate.vector[0];
			def->eval.vector[1] = lex_immediate.vector[1];
			def->eval.vector[2] = lex_immediate.vector[2];
			break;
		case IT_VECTOR4:
			def = Script_CreateTempConstDef( EV_COLOR );
			def->eval.color[0] = lex_immediate.vector[0];
			def->eval.color[1] = lex_immediate.vector[1];
			def->eval.color[2] = lex_immediate.vector[2];
			def->eval.color[3] = lex_immediate.vector[3];
			break;
		default:
			Com_Error( ERR_DROP, "Unknown immediate type" );
			def = NULL;
			break;
		}
		def->flags |= DEF_IMMEDIATE;
		Lex();
		return def;
	}

	if( Lex_Check( "true" ) ) {
		def = Script_CreateTempConstDef( EV_BOOL );
		def->eval.integer = 1;
		def->flags |= DEF_IMMEDIATE;
		return def;
	}

	if( Lex_Check( "false" ) ) {
		def = Script_CreateTempConstDef( EV_BOOL );
		def->eval.integer = 0;
		def->flags |= DEF_IMMEDIATE;
		return def;
	}

	if( Lex_Check( "$" ) ) {
		name = Script_ParseName();
		f = Cmd_FindMacroFunction( name );
		if( f ) {
			def = Script_CreateTempConstDef( EV_MACRO );
			def->eval.macro = f;
		} else {
			def = Script_CreateTempConstDef( EV_CVAR );
			def->eval.cvar = Cvar_Get( name, "", CVAR_USER_CREATED );
		}
		return def;
	}

	name = Script_ParseName();
	if( !( def = Script_FindDef( name ) ) ) {
		Lex_Error( "%s is undefined", name );
	}

	return def;

}

operator_t *Script_FindSuitableOp( const char *string, def_t *a, def_t *b ) {
	operator_t *op;

	for( op = &opcodes[1]; op->name; op++ ) {
		if( !op->string ) {
			continue;
		}
		if( !strcmp( op->string, string ) ) {
			break;
		}
	}

	if( !op->name ) {
		return NULL;
	}

	while( a->type != op->a || ( b && b->type != op->b ) ) {
		op++;
		if( !op->name || strcmp( op->string, string ) ) {
			return NULL;
		}
	}

	return op;
}

def_t *Script_AttemptConversion( def_t *src, def_t *dst ) {
	operator_t *op;

	if( src->type == dst->type ) {
		return src;
	}

	op = Script_FindOpcode( OP_CV_FI );
	if( !op ) {
		Com_Error( ERR_DROP, "OP_CV_FI not found" );
	}

	while( src->type != op->a || dst->type != op->c ) {
		op++;
		if( !op->name || strcmp( op->string, "<CONV>" ) ) {
			return src;
		}
	}
	
	src = Script_Statement( op, src, NULL );
	
	return src;

	
}

def_t *Script_ApplyConversion( def_t *src, evalType_t dstType ) {
	operator_t *op;

	if( src->type == dstType ) {
		return src;
	}

	op = Script_FindOpcode( OP_CV_FI );
	if( !op ) {
		Com_Error( ERR_DROP, "OP_CV_FI not found" );
	}

	while( src->type != op->a || dstType != op->c ) {
		op++;
		if( !op->name || strcmp( op->string, "<CONV>" ) ) {
			Lex_Error( "Cannot convert '%s' to '%s'",
				simpleTypes[src->type].name,
				simpleTypes[dstType].name );
		}
	}
	
	src = Script_Statement( op, src, NULL );
	
	return src;

	
}

/*
================
Script_FunctionCall
================
*/
def_t *Script_FunctionCall( def_t *def ) {
	def_t *src, *dst;
	int currentArg;
	operator_t *op;

	if( def->type != EV_FUNCTION ) {
		Lex_Error( "'%s' is not a function", def->name );
	}

	if( def->counter ) {
		Lex_Error( "Recursive function calls not supported", def->name );
	}

	def->counter = 1;

	currentArg = 0;
	if( !Lex_Check( ")" ) ) {
		do {
			if( currentArg == def->numArgs ) {
				Lex_Error( "Too many arguments" );
			}

			src = Script_Expression( PRIORITY_TOP );
			dst = def->args[currentArg++];
			if( src->type != dst->type ) {
				if( simpleTypes[src->type].sizeRank > simpleTypes[dst->type].sizeRank ) {
					Lex_Error( "Type mismatch for argument %d: "
						"cannot convert '%s' to '%s' (missing a cast)",
						currentArg, simpleTypes[src->type].name, simpleTypes[dst->type].name );
				}
				src = Script_AttemptConversion( src, dst );
			}
			op = Script_FindSuitableOp( "=", dst, src );
			if( !op ) {
				Lex_Error( "Type mismatch for argument %d: "
					"cannot convert '%s' to '%s'",
					currentArg, simpleTypes[src->type].name, simpleTypes[dst->type].name );
			}
			Script_RightStatement( op, dst, src );
		} while( Lex_Check( "," ) );
		Lex_Expect( ")" );
	}

	def->counter = 0;

	if( currentArg != def->numArgs ) {
		Lex_Error( "Insufficient number of arguments" );
	}

	// allocate return def
	dst = Script_CreateTempDef( def->auxtype );

	Script_Statement( Script_FindOpcode( OP_CALL ), def, dst );

	return dst;
}

/*
================
Script_Term
================
*/
def_t *Script_Term( void ) {
	def_t *def;
	operator_t *op;
	evalType_t type;

	if( Lex_Check( "!" ) ) {
		def = Script_Term();
		if( def->type != EV_BOOL ) {
			Lex_Error( "Expected 'bool' type for unary '!'" );
		}
		op = Script_FindOpcode( OP_NOT );
		def = Script_Statement( op, def, NULL );
		return def;
	}

	if( Lex_Check( "-" ) ) {
		def = Script_Term();
		switch( def->type ) {
		case EV_INT:
			op = Script_FindOpcode( OP_NEG_I );
			break;
		case EV_FLOAT:
			op = Script_FindOpcode( OP_NEG_F );
			break;
		case EV_CVAR:
		case EV_MACRO:
			Lex_Error( "Cannot apply unary '-' to '%s', missing a cast",
				simpleTypes[def->type].name );
			op = NULL;
			break;
		default:
			Lex_Error( "Type mismatch for unary '-'" );
			op = NULL;
			break;
		}
		def = Script_Statement( op, def, NULL );
		return def;
	}

	if( Lex_Check( "~" ) ) {
		def = Script_Term();
		switch( def->type ) {
		case EV_INT:
			break;
		case EV_CVAR:
			op = Script_FindOpcode( OP_CV_CVI );
			def = Script_Statement( op, def, NULL );
			break;
		case EV_MACRO:
			op = Script_FindOpcode( OP_CV_MI );
			def = Script_Statement( op, def, NULL );
			break;
		default:
			Lex_Error( "Type mismatch for unary '~'" );
			break;
		}

		op = Script_FindOpcode( OP_BITINV );
		def = Script_Statement( op, def, NULL );
		return def;
	}

	if( Lex_Check( "(" ) ) {
		if( Script_IsSimpleType() ) {
			type = Script_ParseSimpleType();
			Lex_Expect( ")" );
			def = Script_Term();
			def = Script_ApplyConversion( def, type );
			return def;
		}
		def = Script_Expression( PRIORITY_TOP );
		Lex_Expect( ")" );

		return def;
	}

	def = Script_ParseValue();

	return def;
}

#define FIELD_OFFSET( type, f )		((int)&(( type * )0)->f)

/* TODO */
def_t *Script_FindField( def_t *a, const char *name ) {
	def_t *def;
	eval_t eval;

	/*if( a->flags & DEF_IMMEDIATE ) {
		Lex_Error( "Cannot indirect from immediate\n" );
	}*/

	if( a->type == EV_COLOR ) {
		if( !strcmp( name, "r" ) ) {
			eval.offset = FIELD_OFFSET( def_t, eval.color[0] );
		} else if( !strcmp( name, "g" ) ) {
			eval.offset = FIELD_OFFSET( def_t, eval.color[1] );
		} else if( !strcmp( name, "b" ) ) {
			eval.offset = FIELD_OFFSET( def_t, eval.color[2] );
		} else if( !strcmp( name, "a" ) ) {
			eval.offset = FIELD_OFFSET( def_t, eval.color[3] );
		} else {
			Lex_Error( "Unknown field name '%s'", name );
		}
	} else if( a->type == EV_RECT ) {
		if( !strcmp( name, "x" ) ) {
			eval.offset = FIELD_OFFSET( def_t, eval.rect.x );
		} else if( !strcmp( name, "y" ) ) {
			eval.offset = FIELD_OFFSET( def_t, eval.rect.y );
		} else if( !strcmp( name, "w" ) ) {
			eval.offset = FIELD_OFFSET( def_t, eval.rect.width );
		} else if( !strcmp( name, "h" ) ) {
			eval.offset = FIELD_OFFSET( def_t, eval.rect.height );
		} else {
			Lex_Error( "Unknown field name '%s'", name );
		}
	} else {
			Lex_Error( "Unknown type" );
		}

	def = Script_CreateTempDef( EV_FIELD );
	def->eval = eval;

	return def;
}

/*
================
Script_Expression
================
*/
def_t *Script_Expression( int priority ) {
	operator_t *op;
	def_t *a, *b;
	char *string;

	if( priority == PRIORITY_TERM ) {
		return Script_Term();
	}

	a = Script_Expression( priority - 1 );

	while( 1 ) {
		// check for function call
		if( priority == PRIORITY_CALL && Lex_Check( "(" ) ) {
			return Script_FunctionCall( a );
		}

		for( op = &opcodes[1]; op->name; op++ ) {
			if( !op->string ) {
				continue;
			}
			if( op->priority == -1 ) {
				continue;
			}
			if( ( op->priority & ~( RIGHT_FLAG | UNARY_FLAG ) ) != priority ) {
				continue;
			}
			if( !Lex_Check( op->string ) ) {
				continue;
			}

			if( op->string[0] == '.' ) {
				if( lex.tokenType != TT_NAME ) {
					Lex_Error( "Field name expected" );
				}
				b = Script_FindField( a, lex.token );
				Lex();
			} else {
				if( op->priority & UNARY_FLAG ) {
					b = NULL;
				} else if( op->priority & RIGHT_FLAG ) {
					b = Script_Expression( priority );
				} else {
					b = Script_Expression( priority - 1 );
				}
			}

			if( b && a->type != b->type ) {
				if( op->priority & RIGHT_FLAG ) {
					if( simpleTypes[b->type].sizeRank > simpleTypes[a->type].sizeRank ) {
						Lex_Error( "Cannot assign '%s' to '%s', missing a cast",
							simpleTypes[b->type].name, simpleTypes[a->type].name );
					}
					b = Script_AttemptConversion( b, a );
				} else {
					if( simpleTypes[b->type].sizeRank > simpleTypes[a->type].sizeRank ) {
						a = Script_AttemptConversion( a, b );
					} else {
						b = Script_AttemptConversion( b, a );
					}
				}
			}

			string = op->string;
			while( a->type != op->a || ( b && b->type != op->b ) ) {
				op++;
				if( !op->name || strcmp( op->string, string ) ) {
					Lex_Error( "Type mismatch for '%s'", string );
					break;
				}
			}

			if( op->priority & RIGHT_FLAG ) {
				a = Script_RightStatement( op, a, b );
			} else {
				a = Script_Statement( op, a, b );
			}

			break;
		}

		if( !op->name ) {
			break;
		}
	}

	return a;
	
}

#define MAX_LOOP_DEPTH	32

statement_t	*loopStack[MAX_LOOP_DEPTH];
int loopDepth;

#define MAX_TAIL_STATEMENTS		32

/*
================
Script_ParseStatement
================
*/
void Script_ParseStatement( void ) {
	def_t *def;
	statement_t *patch, *jump;
	int offset;
	statement_t tailStatements[MAX_TAIL_STATEMENTS];
	statementBuffer_t tmpbuf;

	patch = NULL;
	if( Lex_Check( "{" ) ) {
		Script_EnterScope();
		while( !Lex_Check( "}" ) ) {
			Script_ParseStatement();
		}
		Script_LeaveScope();
		return;
	}

	if( Lex_Check( ";" ) ) {
		return;
	}

	/* check for declaration */
	if( Script_IsSimpleType() ) {
		Script_ParseDef();
		return;
	}

	if( Lex_Check( "if" ) ) {
		Lex_Expect( "(" );
		def = Script_Expression( PRIORITY_TOP );
		Lex_Expect( ")" );

		if( def->flags & DEF_IMMEDIATE ) {
			/* TODO: eliminate unreachable code */
		}

		if( def->type != EV_BOOL ) {
			Lex_Error( "Expected 'bool' type for 'if'" );
		}
		Script_Statement( Script_FindOpcode( OP_IFNOT ), def, NULL );

		patch = &stbuf->statements[ stbuf->numStatements - 1 ];

		Script_ParseStatement();

		if( Lex_Check( "else" ) ) {
			jump = Script_AllocStatement();
			jump->op = OP_JUMP;

			patch->param = stbuf->numStatements;

			Script_ParseStatement();

			jump->param = stbuf->numStatements;
		} else {
			patch->param = stbuf->numStatements;
		}

		return;
	}

	if( Lex_Check( "while" ) ) {
		if( loopDepth == MAX_LOOP_DEPTH ) {
			Lex_Error( "MAX_LOOP_DEPTH exceeded" );
		}

		/* save offset for jump statement */
		offset = stbuf->numStatements;

		Lex_Expect( "(" );
		def = Script_Expression( PRIORITY_TOP );
		Lex_Expect( ")" );

		if( def->type != EV_BOOL ) {
			Lex_Error( "Expected 'bool' type for 'while'" );
		}
		Script_Statement( Script_FindOpcode( OP_IFNOT ), def, NULL );

		patch = &stbuf->statements[ stbuf->numStatements - 1 ];
		loopStack[loopDepth++] = patch;
		patch->patchNext = NULL;

		Script_ParseStatement();

		jump = Script_AllocStatement();
		jump->op = OP_JUMP;
		jump->param = offset;

		/* patch break statements */
		loopDepth--;
		for( patch = loopStack[loopDepth]; patch; patch = patch->patchNext ) {
			patch->param = stbuf->numStatements;
		}

		return;
	}

	if( Lex_Check( "for" ) ) {
		if( loopDepth == MAX_LOOP_DEPTH ) {
			Lex_Error( "MAX_LOOP_DEPTH exceeded" );
		}

		Lex_Expect( "(" );

		if( !Lex_Check( ";" ) ) {
			Script_Expression( PRIORITY_TOP );
			Lex_Expect( ";" );
		}

		/* save offset for jump statement */
		offset = stbuf->numStatements;

		if( !Lex_Check( ";" ) ) {
			def = Script_Expression( PRIORITY_TOP );
			if( def->type != EV_BOOL ) {
				Lex_Error( "Expected 'bool' type for 'for'" );
			}
			Script_Statement( Script_FindOpcode( OP_IFNOT ), def, NULL );
			patch = &stbuf->statements[ stbuf->numStatements - 1 ];
			patch->patchNext = NULL;
			Lex_Expect( ";" );
		} else {
			patch = NULL;
		}

		loopStack[loopDepth++] = patch;

		/* parse tail statements into temp buffer */
		if( !Lex_Check( ")" ) ) {
			memset( &tailStatements, 0, sizeof( tailStatements ) );
			tmpbuf.maxStatements = MAX_TAIL_STATEMENTS;
			tmpbuf.statements = tailStatements;
			tmpbuf.numStatements = 0;
			stbuf = &tmpbuf;

			Script_Expression( PRIORITY_TOP );
			Lex_Expect( ")" );

			stbuf = &stbuffer;

			Script_ParseStatement();

			if( stbuf->numStatements + tmpbuf.numStatements > stbuf->maxStatements ) {
				Lex_Error( "MAX_STATEMENTS exceeded" );
			}

			/* copy them back */
			memcpy( stbuf->statements + stbuf->numStatements,
				tailStatements, tmpbuf.numStatements * sizeof( statement_t ) );
			stbuf->numStatements += tmpbuf.numStatements;
		} else {
			Script_ParseStatement();
		}

		jump = Script_AllocStatement();
		jump->op = OP_JUMP;
		jump->param = offset;

		/* patch break statements */
		loopDepth--;
		for( patch = loopStack[loopDepth]; patch; patch = patch->patchNext ) {
			patch->param = stbuf->numStatements;
		}

		return;
	}

	if( Lex_Check( "break" ) ) {
		if( !loopDepth ) {
			Lex_Error( "Illegal 'break'" );
		}
		Lex_Expect( ";" );

		/* add this statement to patch list */
		patch = Script_AllocStatement();
		patch->op = OP_JUMP;
		patch->patchNext = loopStack[ loopDepth - 1 ];
		loopStack[ loopDepth - 1 ] = patch;
		return;
	}

	if( Lex_Check( "return" ) ) {
		if( !currentFunction ) {
			Lex_Error( "Illegal 'return'" );
		}
		if( Lex_Check( ";" ) ) {
			if( currentFunction->auxtype != EV_VOID ) {
				Lex_Error( "'%s' should return a value", currentFunction->name );
			}
			Script_Statement( Script_FindOpcode( OP_LEAVE ), currentFunction, NULL );
			return;
		}

		def = Script_Expression( PRIORITY_TOP );
		Lex_Expect( ";" );

		if( def->type != currentFunction->auxtype ) {
			Lex_Error( "Type mismatch for 'return'" );
		}

		Script_Statement( Script_FindOpcode( OP_LEAVE ), currentFunction, def );
		return;
	}

	/* for debugging */
	if( Lex_Check( "__break" ) ) {
		Script_Statement( Script_FindOpcode( OP_BREAK ), NULL, NULL );
		return;
	}

	Script_Expression( PRIORITY_TOP );
	Lex_Expect( ";" );
}

def_t dummyDef;

def_t *Script_FindFunction( char *name ) {
	def_t *def;

	def = Script_FindLocalDef( name, &globalScope );
	if( def == NULL || def->type != EV_FUNCTION ) {
		return NULL;
	}

	return def;
}

/*
================
Script_ParseFunction
================
*/
void Script_ParseFunction( char *name, evalType_t type ) {
	def_t *def;
	int i;
	statement_t *s;

	if( currentFunction ) {
		Lex_Error( "Functions must be global" );
	}

	def = Script_CreateDef( name, EV_FUNCTION );
	def->counter = 1;
	def->auxtype = type;
	def->eval.offset = stbuf->numStatements;

	for( i = 0; i < MAX_ARGS; i++ ) {
		def->args[i] = &dummyDef;
	}

	currentFunction = def;
	Script_EnterScope();

	Script_ParseParameters();

	if( !Lex_Check( ";" ) ) { // check for prototype
		Lex_Expect( "{" );
		while( !Lex_Check( "}" ) ) {
			Script_ParseStatement();
		}

		s = &stbuf->statements[ stbuf->numStatements - 1 ];
		if( s->op != OP_LEAVE ) {
			if( type != EV_VOID ) {
				Lex_Error( "'%s' should return a value", def->name );
			}
			Script_Statement( Script_FindOpcode( OP_LEAVE ), def, NULL );
		}
	}

	Script_LeaveScope();

	def->counter = 0;

	currentFunction = NULL;

}

void Script_ParseEnum( void ) {
	def_t *def, *init;
	uint32 v;
	char *name;

	v = 0;
	Lex_Expect( "{" );
	while( !Lex_Check( "}" ) ) {
		name = Script_ParseName();
		def = Script_FindLocalDef( name, currentScope );
		if( def ) {
			Lex_Error( "Redefenition of '%s'", name );
		}
		def = Script_CreateDef( name, EV_INT );
		def->flags |= DEF_CONST|DEF_IMMEDIATE|DEF_DONTFREE;
		if( Lex_Check( "=" ) ) {
			init = Script_Expression( PRIORITY_TOP );
			if( init->type != EV_INT ) {
				Lex_Error( "Enumerator intitialization by a non-integer" );
			}
			if( !( init->flags & DEF_IMMEDIATE ) ) {
				Lex_Error( "Enumerator intitialization by a non-constant" );
			}
			v = init->eval.integer;
			Script_FreeDef( init );
		}
		def->eval.integer = v;
		if( !Lex_Check( "," ) ) {
			Lex_Expect( "}" );
			break;
		}
		v++;
	}
	
	Lex_Check( ";" ); /* semicolon is optional */
}

/*
================
Script_ParseDef
================
*/
void Script_ParseDef( void ) {
	evalType_t type;
	def_t *def, *init;
	char *name;
	operator_t *op;

	type = Script_ParseSimpleType();

	while( 1 ) {
		name = Script_ParseName();

		def = Script_FindLocalDef( name, currentScope );
		if( def ) {
			Lex_Error( "Redefenition of '%s'", name );
		}

		if( Lex_Check( "(" ) ) {
			Script_ParseFunction( name, type );
			break;
		}

		def = Script_CreateDef( name, type );

		if( Lex_Check( "=" ) ) {
			// initialization
			init = Script_Expression( PRIORITY_TOP );
			if( !currentFunction ) {
				if( !( init->flags & DEF_IMMEDIATE ) ) {
					Lex_Error( "Initialization by a non-constant" );
				}
				//Lex_Error( "Globals initialization not yet supported" );
				def->eval = init->eval;
				Script_FreeDef( init );
			} else {
				op = Script_FindSuitableOp( "=", def, init );
				if( !op ) {
					init = Script_AttemptConversion( init, def );
					op = Script_FindSuitableOp( "=", def, init );
					if( !op ) {
						Lex_Error( "Type mismatch for initializer" );
					}
				}
				Script_RightStatement( op, def, init );
			}
		}

		if( !Lex_Check( "," ) ) {
			Lex_Expect( ";" );
			break;
		}

	}	
}

void Script_CreateBuiltins( builtinFunction_t *funcs ) {
	builtinFunction_t *f;
	def_t *def;
	int i;

	currentScope = &globalScope;
	stbuf = &stbuffer;

	for( f = funcs; f->name; f++ ) {
		def = Script_CreateDef( f->name, EV_FUNCTION );
		def->flags |= DEF_BUILTIN;
		def->auxtype = f->rettype;
		for( i = 0; f->args[i]; i++ ) {
			def->args[i] = Script_CreateTempDef( f->args[i] );
		}
		def->eval.callback = f->callback;
		def->numArgs = i;
	}
}

/*
================
Script_Parse
================
*/
void Script_Parse( void ) {
	while( lex.tokenType != TT_EOF ) {
		if( Lex_Check( "include" ) ) {
			Lex_ExpectImmediate( IT_STRING );
			Lex_Expect( ";" );
			Script_CompileFile( lex_immediate.string );
			continue;
		}

		if( Lex_Check( "enum" ) ) {
			Script_ParseEnum();
			continue;
		}
		Script_ParseDef();
	}
}

/*
================
Script_CompileFile
================
*/
void Script_CompileFile( const char *name ) {
	char *script;
	lexContext_t savedContext;

	if( !FS_LoadFile( name, ( void ** )&script ) ) {
		Lex_Error( "Couldn't load '%s'", name );
	}

	savedContext = lex;
	Lex_BeginParse( script );

	Script_Parse();

	FS_FreeFile( script );

	lex = savedContext;

}




