/*
Copyright (C) 2003-2006 Andrey Nazarov

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "script_local.h"

#define STACK_SIZE		1024

typedef struct {
	int retpc;
	def_t *retval;
} stackEntry_t;

static stackEntry_t	stack[STACK_SIZE];
static stackEntry_t	*opStack;

static int	execCounter;

static void Script_PrintDef( def_t *def ) {
	static const char *boolStrings[] = { "false", "true" };

	if( !def ) {
		Com_Printf( "NULL" );
		return;
	}

	if( def->flags & DEF_IMMEDIATE ) {
		switch( def->type ) {
		case EV_INT:
			Com_Printf( "<%d>", def->eval.integer );
			break;
		case EV_FLOAT:
			Com_Printf( "<%f>", def->eval.value );
			break;
		case EV_STRING:
			Com_Printf( "<%s>", def->eval.string );
			break;
		case EV_BOOL:
			Com_Printf( "<%s>", boolStrings[def->eval.integer & 1] );
			break;
		case EV_COLOR:
			Com_Printf( "<%f %f %f %f>",
				def->eval.color[0],
				def->eval.color[1],
				def->eval.color[2],
				def->eval.color[3] );
			break;
		default:
			Com_Printf( "<unk_%p>", def );
			break;

		}
		return;
	}

	if( def->flags & DEF_TEMP ) {
		Com_Printf( "tmp_%p", def );
		return;
	}

	Com_Printf( "%s", def->name );

}

static void Script_PrintStatement( statement_t *s ) {
	operator_t *op;

	op = Script_FindOpcode( s->op );

	Com_Printf( "%s, ", op->name );
	Script_PrintDef( s->a ); Com_Printf( ", " );
	Script_PrintDef( s->b ); Com_Printf( ", " );
	Script_PrintDef( s->c ); Com_Printf( "\n" );
}

void Script_DumpFunction( def_t *function ) {
	statement_t *s;

	s = &stbuf->statements[function->eval.offset];
	while( 1 ) {
		Script_PrintStatement( s );
		if( s->op == OP_LEAVE ) {
			break;
		}
		s++;
	}
}

/*
================
Script_Execute
================
*/
void Script_Execute( eval_t *ret, def_t *function, ... ) {
	va_list argptr;
	statement_t *s;
	def_t *arg;
	eval_t *ev;
	int i;
	float *fsrc;
	def_t *a, *b, *c;
	int cycleCounter;
	char *string;
	char buffer[16];

	execCounter++;

	// push parameters
	va_start( argptr, function );
	for( i = 0; i < function->numArgs; i++ ) {
		arg = function->args[i];
		ev = &arg->eval;
		switch( arg->type ) {
		case EV_INT:
			ev->integer = va_arg( argptr, int );
			break;
		case EV_FLOAT:
			ev->value = va_arg( argptr, double );
			break;
		case EV_STRING:
			string = va_arg( argptr, char * );
			Q_strncpyz( ev->string, string, MAX_STRING_CHARS );
			break;
		case EV_VECTOR:
			fsrc = va_arg( argptr, float * );
			VectorCopy( fsrc, ev->vector );
			break;
		case EV_COLOR:
			fsrc = va_arg( argptr, float * );
			VectorCopy( fsrc, ev->vector );
			break;
		default:
			Com_Error( ERR_DROP, "Script_Execute: bad argument type: %d", arg->type );
			break;
		}
	}
	va_end( argptr );

	opStack = &stack[0];

	// push return address
	opStack->retpc = 0;
	opStack->retval = NULL;
	opStack++;

	cycleCounter = 1000000;
	s = &stbuf->statements[function->eval.offset];
	while( 1 ) {
		if( --cycleCounter == 0 ) {
			Com_Error( ERR_DROP, "Runaway loop error" );
		}

		a = s->a;
		b = s->b;
		c = s->c;

		//Script_PrintStatement( s );

		switch( s->op ) {
		case OP_NOP:
			break;
		case OP_ENTER:
			break;
		case OP_LEAVE:
			if( opStack == &stack[0] ) {
				Com_Error( ERR_DROP, "Stack underflow" );
			}
			a->counter--;
			opStack--;
			if( opStack->retpc == 0 ) {
				if( b && ret ) {
					*ret = b->eval;
				}
				return;
			}
			if( b && opStack->retval ) {
				opStack->retval->eval = b->eval;
			}
			s = &stbuf->statements[opStack->retpc];
			break;
		case OP_CALL:
			if( a->flags & DEF_BUILTIN ) {
				a->eval.callback( a, b );
				break;
			}	
			// push return address
			if( a->counter == execCounter ) {
				Com_Error( ERR_DROP, "Entered recursively" );
			}
			if( opStack == &stack[STACK_SIZE] ) {
				Com_Error( ERR_DROP, "Stack overflow" );
			}
			a->counter = execCounter;
			opStack->retval = b;
			opStack->retpc = s - stbuf->statements;
			opStack++; 
			s = &stbuf->statements[ a->eval.offset - 1 ];
			break;

		case OP_STORE_I:
		case OP_STORE_B:
			b->eval.integer = a->eval.integer;
			break;
		case OP_STORE_F:
			b->eval.value = a->eval.value;
			break;
		case OP_STORE_V:
			b->eval.vector[0] = a->eval.vector[0];
			b->eval.vector[1] = a->eval.vector[1];
			b->eval.vector[2] = a->eval.vector[2];
			break;
		case OP_STORE_C:
			b->eval.color[0] = a->eval.color[0];
			b->eval.color[1] = a->eval.color[1];
			b->eval.color[2] = a->eval.color[2];
			b->eval.color[3] = a->eval.color[3];
			break;
		case OP_STORE_R:
			b->eval.rect = a->eval.rect;
			break;
		case OP_STORE_S:
			Q_strncpyz( b->eval.string, a->eval.string, MAX_STRING_CHARS );
			break;
		case OP_STORE_P:
			b->eval.pointer = a->eval.pointer;
			break;

		case OP_STOREP_I:
		case OP_STOREP_B:		
			*( int * )b->eval.pointer = a->eval.integer;
			break;
		case OP_STOREP_F:
			*( float * )b->eval.pointer = a->eval.value;
			break;
		case OP_STOREP_V:
			(( float * )b->eval.pointer)[0] = a->eval.vector[0];
			(( float * )b->eval.pointer)[1] = a->eval.vector[1];
			(( float * )b->eval.pointer)[2] = a->eval.vector[2];
			break;
		case OP_STOREP_C:
			(( float * )b->eval.pointer)[0] = a->eval.color[0];
			(( float * )b->eval.pointer)[1] = a->eval.color[1];
			(( float * )b->eval.pointer)[2] = a->eval.color[2];
			(( float * )b->eval.pointer)[3] = a->eval.color[3];
			break;
		case OP_STOREP_S:
			strcpy( *( char ** )b->eval.pointer, a->eval.string );
			break;

		case OP_LOADP_I:
		case OP_LOADP_B:
			c->eval.integer = *( int * )a->eval.pointer;
			break;
		case OP_LOADP_F:
			c->eval.value = *( float * )a->eval.pointer;
			break;
		case OP_LOADP_V:
			a->eval.vector[0] = (( float * )a->eval.pointer)[0];
			a->eval.vector[1] = (( float * )a->eval.pointer)[1];
			a->eval.vector[2] = (( float * )a->eval.pointer)[2];
			break;
		case OP_LOADP_C:
			a->eval.color[0] = (( float * )a->eval.pointer)[0];
			a->eval.color[1] = (( float * )a->eval.pointer)[1];
			a->eval.color[2] = (( float * )a->eval.pointer)[2];
			a->eval.color[3] = (( float * )a->eval.pointer)[3];
			break;
		case OP_LOADP_S:
			/* simply copies string pointer */
			c->eval.string = *( char ** )a->eval.pointer;
			break;

		case OP_INDIRECT:
			c->eval.pointer = ( byte * )a + b->eval.offset;
			break;

		case OP_INDIRECT_P:
			c->eval.pointer = ( byte * )a->eval.pointer + b->eval.offset;
			break;

		case OP_UMUL_F:
			b->eval.value *= a->eval.value;
			break;
		case OP_UDIV_F:
			if( a->eval.value == 0 ) {
				Com_Error( ERR_DROP, "Division by zero" );
			}
			b->eval.value /= a->eval.value;
			break;
		case OP_UADD_F:
			b->eval.value += a->eval.value;
			break;
		case OP_USUB_F:
			b->eval.value -= a->eval.value;
			break;

		case OP_MUL_F:
			c->eval.value = a->eval.value * b->eval.value;
			break;
		case OP_DIV_F:
			if( b->eval.value == 0 ) {
				Com_Error( ERR_DROP, "Division by zero" );
			}
			c->eval.value = a->eval.value / b->eval.value;
			break;
		case OP_ADD_F:
			c->eval.value = a->eval.value + b->eval.value;
			break;
		case OP_SUB_F:
			c->eval.value = a->eval.value - b->eval.value;
			break;

		case OP_EQ_F:
			c->eval.integer = a->eval.value == b->eval.value;
			break;
		case OP_NE_F:
			c->eval.integer = a->eval.value != b->eval.value;
			break;
		case OP_GE_F:
			c->eval.integer = a->eval.value >= b->eval.value;
			break;
		case OP_GT_F:
			c->eval.integer = a->eval.value > b->eval.value;
			break;
		case OP_LE_F:
			c->eval.integer = a->eval.value <= b->eval.value;
			break;
		case OP_LT_F:
			c->eval.integer = a->eval.value < b->eval.value;
			break;

		case OP_NEG_F:
			c->eval.value = -a->eval.value;	
			break;

		case OP_UINC_F:
			b->eval.value++;
			break;
		case OP_UDEC_F:
			b->eval.value--;
			break;

		case OP_MUL_I:
			c->eval.integer = a->eval.integer * b->eval.integer;
			break;
		case OP_DIV_I:
			if( b->eval.integer == 0 ) {
				Com_Error( ERR_DROP, "Division by zero" );
			}
			c->eval.integer = a->eval.integer / b->eval.integer;
			break;
		case OP_MOD_I:
			if( b->eval.integer == 0 ) {
				Com_Error( ERR_DROP, "Division by zero" );
			}
			c->eval.integer = a->eval.integer % b->eval.integer;
			break;
		case OP_ADD_I:
			c->eval.integer = a->eval.integer + b->eval.integer;
			break;
		case OP_SUB_I:
			c->eval.integer = a->eval.integer - b->eval.integer;
			break;

		case OP_EQ_I:
			c->eval.integer = a->eval.integer == b->eval.integer;
			break;
		case OP_NE_I:
			c->eval.integer = a->eval.integer != b->eval.integer;
			break;
		case OP_GE_I:
			c->eval.integer = a->eval.integer >= b->eval.integer;
			break;
		case OP_GT_I:
			c->eval.integer = a->eval.integer > b->eval.integer;
			break;
		case OP_LE_I:
			c->eval.integer = a->eval.integer <= b->eval.integer;
			break;
		case OP_LT_I:
			c->eval.integer = a->eval.integer < b->eval.integer;
			break;

		case OP_BITOR:
			c->eval.integer = a->eval.integer | b->eval.integer;
			break;
		case OP_BITAND:
			c->eval.integer = a->eval.integer & b->eval.integer;
			break;
		case OP_BITXOR:
			c->eval.integer = a->eval.integer ^ b->eval.integer;
			break;

		case OP_NEG_I:
			c->eval.integer = -a->eval.integer;
			break;
		case OP_BITINV:
			c->eval.integer = ~a->eval.integer;
			break;

		case OP_UMUL_I:
			b->eval.integer *= a->eval.integer;
			break;
		case OP_UDIV_I:
			if( a->eval.integer == 0 ) {
				Com_Error( ERR_DROP, "Division by zero" );
			}
			b->eval.integer /= a->eval.integer;
			break;
		case OP_UADD_I:
			b->eval.integer += a->eval.integer;
			break;
		case OP_USUB_I:
			b->eval.integer -= a->eval.integer;
			break;

		case OP_UBITOR:
			b->eval.integer |= a->eval.integer;
			break;
		case OP_UBITAND:
			b->eval.integer &= a->eval.integer;
			break;
		case OP_UBITXOR:
			b->eval.integer ^= a->eval.integer;
			break;

		case OP_UINC_I:
			b->eval.integer++;
			break;
		case OP_UDEC_I:
			b->eval.integer--;
			break;

		case OP_UADD_S:
			Q_strcat( b->eval.string, MAX_STRING_CHARS, a->eval.string );
			break;

		case OP_ADD_S:
			Q_strncpyz( c->eval.string, a->eval.string, MAX_STRING_CHARS );
			Q_strcat( c->eval.string, MAX_STRING_CHARS, b->eval.string );
			break;

		case OP_EQ_S:
			c->eval.integer = strcmp( a->eval.string, b->eval.string ) == 0;
			break;
		case OP_NE_S:
			c->eval.integer = strcmp( a->eval.string, b->eval.string ) != 0;
			break;
		case OP_GE_S:
			c->eval.integer = strcmp( a->eval.string, b->eval.string ) >= 0;
			break;
		case OP_GT_S:
			c->eval.integer = strcmp( a->eval.string, b->eval.string ) > 0;
			break;
		case OP_LE_S:
			c->eval.integer = strcmp( a->eval.string, b->eval.string ) <= 0;
			break;
		case OP_LT_S:
			c->eval.integer = strcmp( a->eval.string, b->eval.string ) < 0;
			break;

		case OP_IFNOT:
			if( !a->eval.integer ) {
				s = &stbuf->statements[ s->param - 1 ];
			}
			break;
		case OP_NOT:
			c->eval.integer = !a->eval.integer;
			break;
		case OP_JUMP:
			s = &stbuf->statements[ s->param - 1 ];
			break;

		case OP_CV_FI:
			c->eval.integer = a->eval.value;
			break;
		case OP_CV_IF:
			c->eval.value = a->eval.integer;
			break;
		case OP_CV_SI:
			c->eval.integer = atoi( a->eval.string );
			break;
		case OP_CV_IS:
			Com_sprintf( c->eval.string, MAX_STRING_CHARS, "%d", a->eval.integer );
			break;
		case OP_CV_SF:
			c->eval.value = atof( a->eval.string );
			break;
		case OP_CV_FS:
			if( a->eval.value == ( int )a->eval.value ) {
				Com_sprintf( c->eval.string, MAX_STRING_CHARS, "%d", ( int )a->eval.value );
			} else {
				Com_sprintf( c->eval.string, MAX_STRING_CHARS, "%f", a->eval.value );
			}
			break;
		case OP_CV_CS:
			Com_sprintf( c->eval.string, MAX_STRING_CHARS, "(%f %f %f %f)",
				a->eval.color[0], a->eval.color[1], a->eval.color[2], a->eval.color[3] );
			break;
		case OP_CV_BS:
			if( a->eval.integer ) {
				strcpy( c->eval.string, "true" );
			} else {
				strcpy( c->eval.string, "false" );
			}
			break;
		case OP_CV_CVI:
			c->eval.integer = a->eval.cvar->integer;
			break;
		case OP_CV_CVF:
			c->eval.value = a->eval.cvar->value;
			break;
		case OP_CV_CVS:
			c->eval.string = a->eval.cvar->string;
			break;
		case OP_CV_MI:
			a->eval.macro( buffer, sizeof( buffer ) );
			c->eval.integer = atoi( buffer );
			break;
		case OP_CV_MF:
			a->eval.macro( buffer, sizeof( buffer ) );
			c->eval.integer = atof( buffer );
			break;
		case OP_CV_MS:
			a->eval.macro( c->eval.string, MAX_STRING_CHARS );
			break;

		default:
			Com_Error( ERR_DROP, "Undefined opcode %d", s->op );
			break;
		}

		s++;
	}
}

